package;

import entity.ECloud;
import entity.EPlayer;
import entity.ESpear;
import flixel.FlxG;
import flixel.FlxSprite;
import flixel.FlxState;
import flixel.addons.nape.FlxNapeSpace;
import flixel.addons.nape.FlxNapeSprite;
import flixel.addons.text.FlxTextField;
import flixel.math.FlxPoint;
import flixel.text.FlxText;
import flixel.ui.FlxButton;
import flixel.util.FlxColor;
import flixel.util.FlxSpriteUtil;
import haxe.xml.Fast;
import motion.Actuate;
import motion.easing.Bounce;
import nape.callbacks.CbEvent;
import nape.callbacks.CbType;
import nape.callbacks.InteractionCallback;
import nape.callbacks.InteractionListener;
import nape.callbacks.InteractionType;
import nape.phys.BodyType;
import nape.phys.Material;
import openfl.Assets;

class PlayState extends FlxState
{
	var playerCharacters:Array<EPlayer>;
	var _baseY:Int = -66;
	var _mouseStart:FlxPoint;
	var _mouseCurrent:FlxPoint;
	var _mousePrevious:FlxPoint;
	var _lineSprite:FlxSprite;
	var IsWinConditionMet:Bool = false;
	var targetAngularVelocity:Float = 1;
	var targetVelocityX:Float = 60;
	var targetVelocityY:Float = 60;

	override public function create():Void
	{
		super.create();

		add(new FlxSprite(0, 0, AssetPaths.bg__png));
		add(new ECloud(AssetPaths.cloud__png));
		
		var btn:FlxButton = new FlxButton(20, 20, "Restart", function(){FlxG.switchState(new PlayState()); });
		btn.label.color = C.BTN_LABEL_COLOR;
		btn.loadGraphic(AssetPaths.button__png, true, 64, 20);
		add(btn);
		
		var txtlvl:FlxText = new FlxText(20, 45, 0, "Level " + C.CURRENT_LEVEL_INDEX + "/10");
		txtlvl.color = C.BTN_LABEL_COLOR;
		add(txtlvl);
		
		FlxNapeSpace.init();
		FlxNapeSpace.space.gravity.setxy(0, 1500);
		FlxNapeSpace.createWalls( -800, -450, 800 * 2, 450 + _baseY);
		
		var lvl:String = C.CURRENT_LEVEL_INDEX > 9 ? "" + C.CURRENT_LEVEL_INDEX : "0" + C.CURRENT_LEVEL_INDEX;
		loadLevel("assets/levels/level" + lvl + ".oel");
		
		FlxNapeSpace.space.listeners.add(new InteractionListener(CbEvent.BEGIN, InteractionType.COLLISION, C.CB_SPEAR, C.CB_BLOCK, onSpearCollides));
		FlxNapeSpace.space.listeners.add(new InteractionListener(CbEvent.BEGIN, InteractionType.COLLISION, C.CB_GOD, CbType.ANY_BODY, onGodCollides));
		
		_mouseStart = new FlxPoint();
		_mouseCurrent = new FlxPoint();
		_mousePrevious = new FlxPoint();
		
		_lineSprite = new FlxSprite(0, 0);
		_lineSprite.makeGraphic(800, 450, FlxColor.TRANSPARENT);
		add(_lineSprite);
		
	}
	
	public function onSpearCollides(callback:InteractionCallback)
	{
		var body = callback.int1.castBody;
		if(body != null) {
			body.cbTypes.remove(C.CB_SPEAR);
			FlxG.camera.shake(0.001, 0.1);
			FlxG.sound.play("assets/sounds/fx_hit_0" + FlxG.random.int(1, 4) + ".ogg", C.SFX_VOLUME * 0.5);
		}
	}

	public function onGodCollides(callback:InteractionCallback)
	{
		var body = callback.int1.castBody;

		if(body != null) {
			if ((body.velocity.x > targetVelocityX / 2 || body.velocity.x < -targetVelocityX / 2) &&
				(body.velocity.y > targetVelocityY / 2 || body.velocity.y < -targetVelocityY / 2)) {
					FlxG.sound.play(AssetPaths.fx_hit_05__ogg, C.SFX_VOLUME);
				}
						
			if(!IsWinConditionMet) {
				var rotation = body.rotation;
				if (rotation != 0) {
					trace(body.angularVel + ", " + body.velocity);
					if ((body.angularVel > targetAngularVelocity || body.angularVel < -targetAngularVelocity) &&
						(body.velocity.x > targetVelocityX || body.velocity.x < -targetVelocityX) &&
						(body.velocity.y > targetVelocityY || body.velocity.y < -targetVelocityY)) {
						createLevelWinScreen();
					}
				}
			}
		}
	}
	
	function createLevelWinScreen() 
	{
		if (IsWinConditionMet) return;
		
		var stripe:FlxSprite = new FlxSprite(-FlxG.width, FlxG.height * 0.5 - 32);
		stripe.makeGraphic(FlxG.width, 64, 0x99000000);
		add(stripe);
		Actuate.tween(stripe, 2, { x: 0 }).ease(Bounce.easeOut);
		
		var _txtNextLevel:String = C.CURRENT_LEVEL_INDEX == 10 ?  "Play again" : "Next Level";
		
		var _txtWin:FlxText = new FlxText(FlxG.width / 2 - 100, -10, 0, "Level complete!", 16);
		_txtWin.color = FlxColor.WHITE;
		add(_txtWin);
		
		var btn:FlxButton = new FlxButton(FlxG.width / 2 + 150, -30, _txtNextLevel, function(){
			C.CURRENT_LEVEL_INDEX++;
			if (C.CURRENT_LEVEL_INDEX > C.MAX_LEVEL_INDEX) {
				C.CURRENT_LEVEL_INDEX = 1;
			}
			FlxG.switchState(new PlayState());
		});
		btn.label.color = C.BTN_LABEL_COLOR;
		btn.loadGraphic(AssetPaths.button__png, true, 64, 20);
		add(btn);
		
		Actuate.tween(_txtWin, 1, { y: FlxG.height / 2 - 10 }).ease(Bounce.easeOut);
		Actuate.tween(btn, 1, { y: FlxG.height / 2 - 10 }).ease(Bounce.easeOut);

		if (C.CURRENT_LEVEL_INDEX == 10) {
			_txtWin.visible = false;
			var _txtThanks:FlxText = new FlxText(FlxG.width / 2 - 100, -10, 300, "You cleared all levels!\nThanks for playing", 16);
			_txtThanks.color = FlxColor.WHITE;
			add(_txtThanks);
			Actuate.tween(_txtThanks, 1, { y: FlxG.height / 2 - 25 }).ease(Bounce.easeOut);
		}
		
		IsWinConditionMet = true;
	}

	function loadLevel(levelName:String) 
	{
		var xml:Xml = Xml.parse(Assets.getText(levelName));
		var fastXml:Fast = new Fast(xml.firstElement());
		playerCharacters = new Array<EPlayer>();
		
		var entities:List<Fast> = fastXml.nodes.resolve("Entity");
		var bricks:List<Fast> = entities.first().nodes.resolve("Brick");
		var concrete:List<Fast> = entities.first().nodes.resolve("Concrete");
		var glass:List<Fast> = entities.first().nodes.resolve("Glass");
		var pinned:List<Fast> = entities.first().nodes.resolve("Pinned");
		var sungod:List<Fast> = entities.first().nodes.resolve("Sungod");
		var playerStart:List<Fast> = entities.first().nodes.resolve("PlayerStart");

		for (brick in bricks)
		{
			createBrick(Std.parseInt(brick.att.x), Std.parseInt(brick.att.y), Std.parseInt(brick.att.width), Std.parseInt(brick.att.height));
		}
		for (block in concrete)
		{
			createConcreteBlock(Std.parseInt(block.att.x), Std.parseInt(block.att.y), Std.parseInt(block.att.width), Std.parseInt(block.att.height));
		}
		for (pane in glass)
		{
			createGlassPane(Std.parseInt(pane.att.x), Std.parseInt(pane.att.y), Std.parseInt(pane.att.width), Std.parseInt(pane.att.height));
		}
		for (pin in pinned)
		{
			createPin(Std.parseInt(pin.att.x), Std.parseInt(pin.att.y), Std.parseInt(pin.att.width), Std.parseInt(pin.att.height));
		}
		for (god in sungod)
		{
			createSunGod(Std.parseInt(god.att.x), Std.parseInt(god.att.y), 48, 48);
		}
		for (pc in playerStart) {
			var p:EPlayer = new EPlayer(Std.parseInt(pc.att.x) + 16, Std.parseInt(pc.att.y) + 16 + _baseY);
			add(p);
			playerCharacters.push(p);
		}
	}
	
	override public function update(elapsed:Float):Void
	{
		super.update(elapsed);
		
		_mouseCurrent.set(FlxG.mouse.x, FlxG.mouse.y);
		
		if (FlxG.mouse.justPressed) {
			_mouseStart.set(FlxG.mouse.x, FlxG.mouse.y);
		}
		if (FlxG.mouse.pressed && (_mouseCurrent != _mousePrevious)) {
			FlxSpriteUtil.fill(_lineSprite, FlxColor.TRANSPARENT);
			FlxSpriteUtil.drawLine(_lineSprite, _mouseStart.x, _mouseStart.y, FlxG.mouse.x, FlxG.mouse.y, { thickness: 1 });
			_mousePrevious.set(FlxG.mouse.x, FlxG.mouse.y);
		}
		if (FlxG.mouse.justReleased) {
			FlxSpriteUtil.fill(_lineSprite, FlxColor.TRANSPARENT);
			var angle:Float = FlxG.mouse.getPosition().angleBetween(_mouseStart);
			var force:Float = FlxG.mouse.getPosition().distanceTo(_mouseStart);
			angle -= 90;
			
			throwProjectiles(force, angle);
		}
	}
	
	function throwProjectiles(force:Float, angle:Float) 
	{
		for (c in playerCharacters) {
			c.Throw();
			createSpear(c.x, c.y, force, angle + c.angle);
		}
	}
	
	function createBrick(x:Int, y:Int, w:Int, h:Int) 
	{
		var brick:FlxNapeSprite = new FlxNapeSprite(x, y, AssetPaths.brick__png);
		brick.createRectangularBody(w, h, BodyType.DYNAMIC);
		brick.setGraphicSize(w, h);
		brick.setPosition(x + w / 2, y + h + _baseY);
		brick.body.setShapeMaterials(Material.steel());
		brick.body.cbTypes.add(C.CB_BLOCK);
		add(brick);
	}
	
	function createConcreteBlock(x:Int, y:Int, w:Int, h:Int)
	{
		var block:FlxNapeSprite = new FlxNapeSprite(x, y, AssetPaths.concrete__png);
		block.createRectangularBody(w, h, BodyType.DYNAMIC);
		block.setGraphicSize(w, h);
		block.setPosition(x + w / 2, y + h + _baseY);
		block.body.setShapeMaterials(Material.steel());
		block.body.mass *= 1.5;
		block.body.cbTypes.add(C.CB_BLOCK);
		add(block);
	}
	
	function createGlassPane(x:Int, y:Int, w:Int, h:Int)
	{
		var pane:FlxNapeSprite = new FlxNapeSprite(x, y, AssetPaths.glass__png);
		pane.createRectangularBody(w, h, BodyType.DYNAMIC);
		pane.setGraphicSize(w, h);
		pane.setPosition(x + w / 2, y + h + _baseY);
		pane.body.setShapeMaterials(Material.glass());
		pane.body.cbTypes.add(C.CB_BLOCK);
		add(pane);
	}
	
	function createPin(x:Int, y:Int, w:Int, h:Int)
	{
		var pin:FlxNapeSprite = new FlxNapeSprite(x, y, AssetPaths.pin__png, false);
		pin.createCircularBody(w / 2, BodyType.DYNAMIC);
		pin.setGraphicSize(w, h);
		pin.setPosition(x + (w * 0.5), y + (h * 0.65) + _baseY);
		pin.body.cbTypes.add(C.CB_BLOCK);
		add(pin);
		pin.body.type = BodyType.STATIC;
	}
	
	function createSunGod(x:Int, y:Int, w:Int, h:Int)
	{
		var god:FlxNapeSprite = new FlxNapeSprite(x, y, AssetPaths.sungod__png, false);
		god.createCircularBody(w / 2, BodyType.DYNAMIC);
		god.setGraphicSize(w, h);
		god.setPosition(x + w / 2, y + h + _baseY);
		god.body.setShapeMaterials(Material.steel());
		god.body.cbTypes.add(C.CB_BLOCK);
		god.body.cbTypes.add(C.CB_GOD);
		add(god);
	}
	
	function createSpear(x:Float, y:Float, force:Float, angle:Float) 
	{
		var spear:ESpear = new ESpear(x, y, AssetPaths.spear__png, force, angle);
		
		add(spear);
	}
}
